import torch
from torch import Tensor
import torch.nn.functional as F
from typing import List, Tuple, Optional, Union
import torch.nn as nn
import math

# @torch.jit.script
def drop_path(x: Tensor, drop_prob: float = 0., training: bool = False) -> Tensor:
    """Drop paths (Stochastic Depth) per sample (when applied in main path of residual blocks).

    TIMM

    This is the same as the DropConnect impl I created for EfficientNet, etc networks, however,
    the original name is misleading as 'Drop Connect' is a different form of dropout in a separate paper...
    See discussion: https://github.com/tensorflow/tpu/issues/494#issuecomment-532968956 ... I've opted for
    changing the layer and argument names to 'drop path' rather than mix DropConnect as a layer name and use
    'survival rate' as the argument.

    """
    if drop_prob == 0. or not training:
        return x

    keep_prob = 1 - drop_prob
    shape = (x.shape[0],) + (1,) * (x.ndim - 1)  # work with diff dim tensors, not just 2D ConvNets
    random_tensor = keep_prob + torch.rand(shape, dtype=x.dtype, device=x.device)
    random_tensor.floor_()  # binarize
    output = x.div(keep_prob) * random_tensor
    return output


class DropPath(nn.Module):
    """Drop paths (Stochastic Depth) per sample  (when applied in main path of residual blocks).
    """

    def __init__(self, drop_prob=None):
        super(DropPath, self).__init__()
        self.drop_prob = drop_prob

    def forward(self, x):
        return drop_path(x, self.drop_prob, self.training)


class Block(nn.Module):
    r""" ConvNeXt Block. There are two equivalent implementations:
    (1) DwConv -> LayerNorm (channels_first) -> 1x1 Conv -> GELU -> 1x1 Conv; all in (N, C, H, W)
    (2) DwConv -> Permute to (N, H, W, C); LayerNorm (channels_last) -> Linear -> GELU -> Linear; Permute back
    We use (2) as we find it slightly faster in PyTorch

    Args:
        dim (int): Number of input channels.
        drop_path (float): Stochastic depth rate. Default: 0.0
        layer_scale_init_value (float): Init value for Layer Scale. Default: 1e-6.
    """

    def __init__(self, dim: int, drop_path: Optional[float] = 0.0, kernel_size: int = 7, dilation: int = 1,
                 layer_scale_init_value: Optional[float] = 1e-2, activation: Optional = None) -> None:
        super().__init__()

        padding = kernel_size // 2
        kernel_size = (kernel_size, ) * 3 if isinstance(kernel_size, int) else kernel_size

        self.dwconv = nn.Conv3d(dim, dim, kernel_size=kernel_size, padding=padding, groups=dim, dilation=dilation)  # depthwise conv
        self.norm = LayerNorm(dim, eps=1e-6)
        self.pwconv1 = nn.Linear(dim, 4 * dim)  # pointwise/1x1 convs, implemented with linear layers
        self.act = activation() if activation else nn.GELU()
        self.pwconv2 = nn.Linear(4 * dim, dim)
        self.gamma = nn.Parameter(layer_scale_init_value * torch.ones((dim)),
                                  requires_grad=True) if layer_scale_init_value > 0 else None
        self.drop_path: Union[DropPath, nn.Identity] = DropPath(drop_path) if drop_path > 0. else nn.Identity()

        self.dim = dim

    def __repr__(self):
        out = f'Block[In={self.dim}, Out={self.dim}]' \
              f'\n\tDepthWiseConv[In={self.dwconv.in_channels}, Out={self.dwconv.out_channels}, KernelSize={self.dwconv.kernel_size}]' \
              f'\n\tLayerNorm[Shape={self.norm.normalized_shape}]' \
              f'\n\tPointWiseConv[In={self.pwconv1.in_features}, Out={self.pwconv1.out_features}]' \
              f'\n\tGELU[]' \
              f'\n\tPointWiseConv[In={self.pwconv2.in_features}, Out={self.pwconv2.out_features}]' \
              f'\n\tScale[]' \
              f'\n\tDropPath[]'

        return out

    def forward(self, x: Tensor) -> Tensor:  # Image of shape (B, C, H, W) -> (10, 3, 100, 100)
        input = x
        x = self.dwconv(x)
        x = x.permute(0, 2, 3, 4, 1)  # (N, C, H, W) -> (N, H, W, C)
        x = self.norm(x)
        x = self.pwconv1(x)
        x = self.act(x)
        x = self.pwconv2(x)
        if self.gamma is not None:
            x = self.gamma * x
        x = x.permute(0, 4, 1, 2, 3)  # (N, H, W, C) -> (N, C, H, W)

        x = input + self.drop_path(x)
        return x


class LayerNorm(nn.Module):
    r""" LayerNorm that supports two data formats: channels_last (default) or channels_first.
    The ordering of the dimensions in the inputs. channels_last corresponds to inputs with
    shape (batch_size, height, width, channels) while channels_first corresponds to inputs
    with shape (batch_size, channels, height, width).
    """

    def __init__(self, num_features, eps=1e-6, data_format="channels_last"):
        super().__init__()
        self.weight = nn.Parameter(torch.ones(num_features))
        self.bias = nn.Parameter(torch.zeros(num_features))
        self.eps = eps
        self.data_format = data_format
        if self.data_format not in ["channels_last", "channels_first"]:
            raise NotImplementedError
        self.normalized_shape = (num_features,)

        # small optimization. Avoids repeated branching in forward call...
        self.norm_functon = F.layer_norm if self.data_format == 'channels_last' else self.layer_norm_channels_fist

    def forward(self, x: Tensor) -> Tensor:
        # if self.data_format == "channels_last":
        #     x = F.layer_norm(x, self.normalized_shape, self.weight, self.bias, self.eps)
        # elif self.data_format == "channels_first":
        #     # u = x.mean(1, keepdim=True)
        #     # s = (x - u).pow(2).mean(1, keepdim=True)
        #     # x = (x - u) / torch.sqrt(s + self.eps)
        #     # x = self.weight.view(1, -1, 1, 1, 1) * x + self.bias.view(1, -1, 1, 1, 1)
        #     x = self.layer_norm_channels_fist(x, self.weight, self.bias, self.eps)

        return self.norm_functon(x, self.normalized_shape, self.weight, self.bias, self.eps)

    @staticmethod
    # @torch.jit.script  # Fuze operation to potentially take place on a single cuda kernel
    def layer_norm_channels_fist(x: Tensor, shape: Tuple[int], weight: Tensor, bias: Tensor, eps: float) -> Tensor:
        u = x.mean(1, keepdim=True)
        s = (x - u).pow(2).mean(1, keepdim=True)
        x = (x - u) / torch.sqrt(s + eps)
        return weight.view(1, -1, 1, 1, 1) * x + bias.view(1, -1, 1, 1, 1)


class UpSampleLayer(nn.Module):
    def __init__(self, in_channels: int, out_channels: int, method='nearest'):
        super(UpSampleLayer, self).__init__()
        self.method = method
        self.norm = LayerNorm(in_channels, eps=1e-6, data_format="channels_first")
        self.compress = nn.Conv3d(in_channels, out_channels, kernel_size=(3, 3, 3), stride=1, padding=1)

    def forward(self, x: Tensor, shape: List[int]) -> Tensor:
        'Upsample layer'
        # print(x.shape ,x.numel(), x.max(), x.min())
        x = F.interpolate(x, shape[2::], mode=self.method)
        x = self.norm(x)
        x = self.compress(x)
        return x


class ConcatConv(nn.Module):
    def __init__(self, in_channels: int):
        super(ConcatConv, self).__init__()
        self.conv = nn.Conv3d(in_channels * 2, in_channels, kernel_size=1, stride=1, padding=0)

    def forward(self, x: Tensor, y: Tensor) -> Tensor:
        # print(f'ConcatConv Called: x: {x.shape}, y: {y.shape}')
        x = torch.cat((x, y), dim=1)
        x = self.conv(x)
        return x


class r_block(nn.Module):
    """
    Recurrently applies a block to an input and outputs later.
    """

    def __init__(self, in_channels: int, out_channels: int, n_loop: int, dilation: Tuple[int] = (1, 1, 1),
                 padding: int = 3):
        super(r_block, self).__init__()

        self.n = n_loop

        self.conv1x1 = nn.Conv3d(in_channels * 2, in_channels, kernel_size=(1, 1, 1))
        self.conv7x7_1 = nn.Conv3d(in_channels, in_channels, kernel_size=(7, 7, 7), padding=padding, stride=(1, 1, 1),
                                   dilation=dilation)
        self.conv7x7_2 = nn.Conv3d(in_channels, out_channels, kernel_size=(7, 7, 7), padding=padding, stride=(1, 1, 1),
                                   dilation=dilation)

        self.activation = nn.LeakyReLU()

        self.batch_norm_conv_1x1 = nn.BatchNorm3d(in_channels)
        self.batch_norm_conv_7x7_1 = nn.BatchNorm3d(in_channels)
        self.batch_norm_conv_7x7_2 = nn.BatchNorm3d(out_channels)

    def forward(self, x: Tensor) -> Tensor:
        y = torch.zeros(x.shape, device=x.device)
        for _ in range(self.n):
            y = torch.cat((x, y), dim=1)
            y = self.activation(self.batch_norm_conv_1x1(self.conv1x1(y)))
            # y = self.activation(self.batch_norm_conv_7x7_3(self.conv7x7_3(y)))
            y = self.activation(self.batch_norm_conv_7x7_1(self.conv7x7_1(y)))

        y = self.activation(self.batch_norm_conv_7x7_2(self.conv7x7_2(y)))

        return y


def _no_grad_trunc_normal_(tensor, mean, std, a, b):
    # Cut & paste from PyTorch official master until it's in a few official releases - RW
    # Method based on https://people.sc.fsu.edu/~jburkardt/presentations/truncated_normal.pdf
    def norm_cdf(x):
        # Computes standard normal cumulative distribution function
        return (1. + math.erf(x / math.sqrt(2.))) / 2.

    if (mean < a - 2 * std) or (mean > b + 2 * std):
        warnings.warn("mean is more than 2 std from [a, b] in nn.init.trunc_normal_. "
                      "The distribution of values may be incorrect.",
                      stacklevel=2)

    with torch.no_grad():
        # Values are generated by using a truncated uniform distribution and
        # then using the inverse CDF for the normal distribution.
        # Get upper and lower cdf values
        l = norm_cdf((a - mean) / std)
        u = norm_cdf((b - mean) / std)

        # Uniformly fill tensor with values from [l, u], then translate to
        # [2l-1, 2u-1].
        tensor.uniform_(2 * l - 1, 2 * u - 1)

        # Use inverse cdf transform for normal distribution to get truncated
        # standard normal
        tensor.erfinv_()

        # Transform to proper mean, std
        tensor.mul_(std * math.sqrt(2.))
        tensor.add_(mean)

        # Clamp to ensure it's in the proper range
        tensor.clamp_(min=a, max=b)
        return tensor


def trunc_normal_(tensor, mean=0., std=1., a=-2., b=2.):
    # type: (Tensor, float, float, float, float) -> Tensor
    r"""Fills the input Tensor with values drawn from a truncated
    normal distribution. The values are effectively drawn from the
    normal distribution :math:`\mathcal{N}(\text{mean}, \text{std}^2)`
    with values outside :math:`[a, b]` redrawn until they are within
    the bounds. The method used for generating the random values works
    best when :math:`a \leq \text{mean} \leq b`.
    Args:
        tensor: an n-dimensional `torch.Tensor`
        mean: the mean of the normal distribution
        std: the standard deviation of the normal distribution
        a: the minimum cutoff value
        b: the maximum cutoff value
    Examples:
        >>> w = torch.empty(3, 5)
        >>> nn.init.trunc_normal_(w)
    """
    return _no_grad_trunc_normal_(tensor, mean, std, a, b)



import torch
from torch import Tensor
import torchvision.ops
from typing import Union, Optional, Tuple, Dict

Device = Union[str, torch.device]


@torch.jit.script
def square_boxes(boxes: Tensor, size: int = 10) -> Tensor:
    boxes = torchvision.ops.box_convert(boxes, 'xyxy', 'cxcywh')
    boxes[:, [2, 3]] = size
    boxes = torchvision.ops.box_convert(boxes, 'cxcywh', 'xyxy')
    return boxes


@torch.jit.script
def calculate_accuracy(ground_truth: Dict[str, Tensor], predictions: Dict[str, Tensor],
                       device: Optional[Device] = None, threshold: float = 0.1) -> Tuple[Tensor]:

    device = device if device else ground_truth['boxes'].device

    # We want to set each box as a perfect square
    _gt = square_boxes(ground_truth['boxes'].to(device), size=10)
    _pred = square_boxes(predictions['boxes'].to(device), size=10)

    iou = torchvision.ops.box_iou(_gt, _pred)

    # GT: indices of GT boxes with CORRECT predictions
    gt_max, gt_indicies = iou.max(dim=1)
    gt = torch.logical_not(gt_max.gt(threshold)) if iou.shape[1] > 0 else torch.ones(0)

    # Pred: indices of prediction boxes with associated GT box
    pred = torch.logical_not(iou.max(dim=0)[0].gt(threshold)) if iou.shape[0] > 0 else torch.ones(0)

    true_positive = torch.sum(torch.logical_not(gt))
    false_positive = torch.sum(pred)
    false_negative = torch.sum(gt)

    tempmat = iou[gt.eq(0), :]
    if tempmat.numel() > 0:
        associated_box_ind = tempmat.argmax(1)
        # of all gt cells with a prediction, what percentage of them have the correct class label?
        cls_acc = (ground_truth['labels'][gt.eq(0)] == pred['labels'][associated_box_ind]).sum() / len(
            ground_truth['labels'][gt.eq(0)])
        ihc = (ground_truth['labels'][gt.eq(0)] == 1).int()
        ihc_acc = (ihc == pred['labels'][associated_box_ind]).sum() / ihc.gt(0).sum()
        ohc = (ground_truth['labels'][gt.eq(0)] == 2).int()
        ohc_acc = (ohc == pred['labels'][associated_box_ind]).sum() / ohc.gt(0).sum()
    else:
        ihc_acc = 0
        ohc_acc = 0
        cls_acc = 0


    return true_positive, false_positive, false_negative, cls_acc





import torch
from torch import Tensor

import hcat.lib.functional
from hcat.lib.functional import IntensityCellReject
from hcat.backends.backend import Backend
from hcat.models.r_unet import embed_model as RUnet
from hcat.models.unext import UNeXT
from hcat.train.transforms import median_filter, erosion
import hcat.lib.utils

from typing import Dict, Optional

import matplotlib.pyplot as plt
from torch.utils.tensorboard import SummaryWriter

from torchvision.utils import draw_keypoints, flow_to_image, make_grid
def write_progress(writer,  tag, epoch, images, probability_map, embedding, vector, centroids):
    keypoint = centroids[0, :, [1, 0]].unsqueeze(0)

    _a = images[0, [0, 0, 0], :, :, 15].cpu()
    _b = probability_map[0, [0, 0, 0], :, :, 15].gt(0.5).float().cpu()


    _overlay = hcat.lib.utils.make_embedding_image(embedding).unsqueeze(0)[[0,0,0],...]
    _overlay = _overlay.div(20).clamp(0, 1)

    _c = draw_keypoints(_overlay.mul(255).round().type(torch.uint8),
                        keypoint, colors='red', radius=3).cpu()
    _d = draw_keypoints(flow_to_image(vector[0, [1, 0], :, :, 15].float()),
                        keypoint, colors='red', radius=3).cpu()

    img_list = [_a, _b, _c, _d]
    for i, im in enumerate(img_list):
        assert isinstance(im, torch.Tensor), f'im {i} is not a Tensor instead it is a {type(im)}, {img_list[i]}'
        assert img_list[0].shape == img_list[
            i].shape, f'im {i} is has shape {im.shape}, not {img_list[0].shape}'

    _img = make_grid(img_list, nrow=1, normalize=True, scale_each=True)

    writer.add_image(tag, _img, epoch, dataformats='CWH')

class SpatialEmbedding(Backend):
    def __init__(self,
                 sigma: Optional[Tensor] = torch.tensor([2, 2, 2]),
                 device: Optional[str] = 'cuda',
                 model_loc: Optional[str] = None,
                 postprocessing: Optional[bool] = True,
                 scale: Optional[int] = 25,
                 figure: Optional[str] = None,
                 architecture: Optional[RUnet] = UNeXT):
        """
        Initialize Spatial embedding Algorithm.

        :param sigma: Tensor[sigma_x, sigma_y, sigma_z] values for gaussian probability estimation.
        :param device: String value for torch device by which to run segmentation backbone on.
        :param model_loc: Path to trained model files.
        :param postprocessing: Disable segmentation postprocessing, namely
        :param scale: scale factor based on max diameter of object
        :param figure: filename and path of diagnostic figure which may be rendered
        """

        super(SpatialEmbedding, self).__init__()

        # self.url = 'https://github.com/buswinka/hcat/blob/master/modelfiles/spatial_embedding.trch?raw=true'
        self.url = None
        self.scale = torch.tensor(scale)
        self.device = device
        self.sigma = sigma.to(device)
        self.postprocessing = postprocessing

        self.figure = figure
        self.writer = SummaryWriter()

        if self.url:
            self.model = self._model_loader_url(self.url, architecture, device)
        else:
            self.model = self._model_loader_path(model_loc, architecture, device)

        self.vector_to_embedding = torch.jit.script(
            hcat.lib.functional.VectorToEmbedding(scale=self.scale).requires_grad_(False).eval())

        self.embedding_to_probability = hcat.lib.functional.EmbeddingToProbability(scale=self.scale).requires_grad_(False).eval()

        self.estimate_centroids = hcat.lib.functional.EstimateCentroids(scale=self.scale,
                                                                        n_erode=2,
                                                                        min_samples=80,
                                                                        ).requires_grad_(False)

        self.filter = median_filter(kernel_targets=3, rate=1, device=device)
        self.binary_erosion = erosion(device=device)

        self.intensity_rejection = IntensityCellReject()

        self.nms = hcat.lib.functional.nms().requires_grad_(False)

        self.centroids = None
        self.vec = None
        self.embed = None
        self.prob = None
        self.e = 0

    def forward(self, image: Tensor) -> Tensor:
        """
        Inputs an image and outputs a probability mask of everything seen in the image.

        .. note::
           Call the module as a function to execute this method (similar to torch.nn.module).

        .. warning:
           Will not raise an error upon failure, instead returns None and prints to standard out

        Example:

        >>> from hcat.backends.spatial_embedding import SpatialEmbedding
        >>> import torch
        >>> backend = SpatialEmbedding()
        >>> image = torch.load('path/to/my/image.trch')
        >>> assert image.ndim == 5 # Shape should be [B, C, X, Y, Z]
        >>> masks = backend(image)

        :param image: [B, C=4, X, Y, Z] input image
        :return: [B, 1, X, Y, Z] output segmentation mask where each pixel value is a cell id (0 is background)
        """
        assert image.ndim == 5
        assert image.shape[1] == 1
        assert image.min() >= -1
        assert image.max() <= 1

        image = self.filter(image.to(self.device))
        image = image.to(self.device)
        b, c, x, y, z = image.shape

        if self.image_reject and self._is_image_bad(image):
            return torch.zeros((b, 0, x, y, z), device=self.device)


        out: Tensor = self.model(image) # Evaluate Neural Network Model

        # Assign Outputs
        probability_map = out[:, [-1], ...]
        out = out[:, 0:3:1, ...]

        self.prob = probability_map.cpu()
        self.vec = out.cpu()

        out: Tensor = self.vector_to_embedding(out, n=1)
        self.embed = out.cpu()

        self.centroids: Dict[str, Tensor] = self.estimate_centroids(self.vector_to_embedding(self.vec.cuda(), n=10), probability_map)

        out: Tensor = self.embedding_to_probability(out, self.centroids, self.sigma)

        # write_progress(self.writer, 'test', self.e, image, probability_map.gt(0.5).float(), self.embed, vector=self.vec, centroids=self.centroids)

        self.e += 1

        # # Reject cell masks that overlap or meet min Myo7a criteria
        # if self.postprocessing:
        #     out: Tensor = self.intensity_rejection(out, image)

        if out.numel() == 0:
            return torch.zeros((b, 0, x, y, z), device=self.device)

        ind = self.nms(out, 0.5)
        out = out[:, ind, ...]

        # Take probabilities and generate masks!
        probability_map = probability_map.lt(0.8).squeeze(1)
        for i in range(out.shape[1]):
            out[:, i, ...][probability_map] = 0

        self.zero_grad()

        return out



import torch
from torch import Tensor
import torchvision.transforms.functional as ttf
from typing import Dict, Tuple, Union, Sequence, List

"""
dl = detection_dataset(path=['data/detection/train', './data/detection/train_external'],
                                  batch_size=32,
                                  simple_class=True,
                                  transforms=torchvision.transforms.Compose([
                                      t.random_crop(shape=(300, 300, 16), box_window=True),
                                      t.to_cuda(),
                                      t.affine3d(rate=0.33, scale=(0.85, 1.2)),  # broken 2d
                                      t.elastic_deformation(grid_shape=(15, 15), scale=1, rate=0.2),  # broken 2d
                                      t.random_h_flip(rate=0.5),
                                      t.random_v_flip(rate=0.5),
                                      t.gaussian_blur(rate=0.1, kernel_targets=torch.tensor([3])),
                                      t.adjust_brightness(range_brightness=(-0.1, 0.1), rate=0.33),
                                      t.random_noise(gamma=0.05, rate=0.2),
                                  ]))
"""


@torch.jit.script
def _get_box(mask: Tensor, device: str, threshold: int) -> Tuple[Tensor, Tensor]:
    # mask in shape of 300, 400, 1 [H, W, z=1]
    nonzero = torch.nonzero(mask)  # Q, 3=[x,y,z]
    label = mask.max()

    box = torch.tensor([-1, -1, -1, -1], dtype=torch.long, device=device)

    # Recall, image in shape of [C, H, W]

    if nonzero.numel() > threshold:
        x0 = torch.min(nonzero[:, 1])
        x1 = torch.max(nonzero[:, 1])
        y0 = torch.min(nonzero[:, 0])
        y1 = torch.max(nonzero[:, 0])

        if (x1 - x0 > 0) and (y1 - y0 > 0):
            box[0] = x0
            box[1] = y0
            box[2] = x1
            box[3] = y1

    return label, box


@torch.jit.script
def _remove_bad_masks(image: Tensor, min_cell_volume: int, device: str) -> Tensor:
    """
    Input an boolean image volume of shape [X, Y, Z] and determine the number of nonzero pixels.

    Sets the segmentation mask of a cell to ZERO if both conditions are met:
        1 - The cell is touching an edge (All edges of volume are considered)
        2 - The cell is smaller than a lower bound set by arg: min_cell_volume

    :param image: [X, Y, Z]
    :param min_cell_volume: min number of voxels needed to keep a cell mask as valid

    :return: torch.Tensor of shape [X, Y, Z]. Returns image if no cells are touching the edge.
    """
    ind = torch.nonzero(image)  # Determine location of nonzero values -> torch.tensor([:, [X, Y, Z]])

    # Loop over each edge (i) and check if the mask is touching.
    # Implicitly, if a nonzero value is zero it is on an edge.
    for i in range(3):
        remove_bool = torch.any(ind[:, i] == 0) or torch.any(ind[:, i] == image.shape[i] - 1)
        remove_bool = remove_bool if image.nonzero().shape[0] < min_cell_volume else False

        # Remove cell if it touches the edge and is small. No need for further computation.
        if remove_bool:
            image = torch.zeros(image.shape, device=device)
            break

    return image


@torch.jit.script
def merged_transform_2D(data_dict: Dict[str, Tensor]) -> Dict[str, Tensor]:
    with torch.no_grad():
        # CONSTANTS #
        DEVICE: str = str(data_dict['image'].device)

        # Image should be in shape of [C, H, W]
        CROP_WIDTH = torch.tensor([300], device=DEVICE)
        CROP_HEIGHT = torch.tensor([300], device=DEVICE)

        AFFINE_RATE = torch.tensor(-1, device=DEVICE)
        AFFINE_SCALE = torch.tensor((0.75, 1.25), device=DEVICE)
        AFFINE_YAW = torch.tensor((-180, 180), device=DEVICE)
        AFFINE_SHEAR = torch.tensor((-4, 4), device=DEVICE)

        FLIP_RATE = torch.tensor(0.5, device=DEVICE)

        BLUR_RATE = torch.tensor(0.3, device=DEVICE)
        BLUR_KERNEL_TARGETS = torch.tensor([3], device=DEVICE, dtype=torch.int)

        BRIGHTNESS_RATE = torch.tensor(0.5, device=DEVICE)
        BRIGHTNESS_RANGE_GFP = torch.tensor((-0.3, 0.05), device=DEVICE)
        BRIGHTNESS_RANGE_ACTIN = torch.tensor((-0.1, 0.125), device=DEVICE)

        CONTRAST_RATE = torch.tensor(0.3, device=DEVICE)
        CONTRAST_RANGE = torch.tensor((0.75, 2.), device=DEVICE)

        NOISE_GAMMA = torch.tensor(0.07, device=DEVICE)
        NOISE_RATE = torch.tensor(0.5, device=DEVICE)

        image: Tensor = data_dict['image'].to(DEVICE)
        boxes: Tensor = data_dict['boxes'].to(DEVICE)
        labels = data_dict['labels'].to(DEVICE)

        # ---------- Random Crop
        w = CROP_WIDTH
        h = CROP_HEIGHT

        ind = torch.randint(boxes.shape[0], (1, 1), dtype=torch.long, device=DEVICE)  # randomly select box

        box = boxes[ind, :].squeeze()
        x0 = box[0]
        y0 = box[1]

        x0 = x0.sub(torch.floor(w / 2)).long()
        y0 = y0.sub(torch.floor(h / 2)).long()
        x1 = x0 + w
        y1 = y0 + h

        image = image[:, y0.item():y1.item(), x0.item():x1.item(), :]

        # Image should be of shape [C, H, W]

        ind_x = torch.logical_and(boxes[:, 0] < x1, boxes[:, 2] > x0)
        ind_y = torch.logical_and(boxes[:, 1] < y1, boxes[:, 3] > y0)
        ind = torch.logical_and(ind_x, ind_y)
        boxes = boxes[ind, :]

        labels = labels[ind]

        boxes[:, 0] -= x0
        boxes[:, 1] -= y0
        boxes[:, 2] -= x0
        boxes[:, 3] -= y0

        boxes[:, 0] = torch.clamp(boxes[:, 0], 0, w.item())
        boxes[:, 1] = torch.clamp(boxes[:, 1], 0, h.item())
        boxes[:, 2] = torch.clamp(boxes[:, 2], 0, w.item())
        boxes[:, 3] = torch.clamp(boxes[:, 3], 0, h.item())

        # -------------------affine
        if torch.rand(1, device=DEVICE) < AFFINE_RATE:

            ## --------------- box 2 mask
            _, x, y, z = image.shape
            n = boxes.shape[0]

            mask = torch.zeros((n, x, y, z), device=DEVICE, dtype=torch.int)
            for i in range(n):
                x0 = boxes[i, 0]
                y0 = boxes[i, 1]
                x1 = boxes[i, 2]
                y1 = boxes[i, 3]

                mask[i, y0:y1, x0:x1, :] = labels[i]  # class label

            angle = (AFFINE_YAW[0] - AFFINE_YAW[1]) * torch.rand(1, device=DEVICE) + AFFINE_YAW[0]
            shear = (AFFINE_SHEAR[0] - AFFINE_SHEAR[1]) * torch.rand(1, device=DEVICE) + AFFINE_SHEAR[0]
            scale = (AFFINE_SCALE[0] - AFFINE_SCALE[1]) * torch.rand(1, device=DEVICE) + AFFINE_SCALE[0]

            image = ttf.affine(image[..., 0], angle=angle.item(), shear=[float(shear.item())], scale=scale.item(),
                               translate=[0, 0]).unsqueeze(-1)
            mask = ttf.affine(mask[..., 0], angle=angle.item(), shear=[float(shear.item())], scale=scale.item(),
                              translate=[0, 0]).unsqueeze(-1)

            ## ------------------- Corrections
            if mask.shape[0] == 0: print(f'{mask.shape}')

            processed_masks: List[torch.jit.Future[Tensor]] = []

            for i in range(mask.shape[0]):
                processed_masks.append(torch.jit.fork(_remove_bad_masks, mask[i, ...], 10, DEVICE))

            results: List[Tensor] = []
            for future in processed_masks:
                results.append(torch.jit.wait(future))
            mask = torch.stack(results)

            ## ------------------- Mask to Box
            _, x, y, z = image.shape
            n = mask.shape[0]

            boxes = torch.zeros([n, 4], device=DEVICE, dtype=torch.long)
            labels = torch.zeros([n], device=DEVICE, dtype=torch.long)

            processed_boxes: List[torch.jit.Future[Tuple[Tensor, Tensor]]] = []

            for i in range(n):
                # _get_box returns Tuple[Tensor (labels.shape=[N]), Tensor (boxes.shape=[N,4])]
                processed_boxes.append(torch.jit.fork(_get_box, mask=mask[i, ...], device=DEVICE, threshold=20))

            results: List[Tuple[Tensor, Tensor]] = []
            for future in processed_boxes:
                results.append(torch.jit.wait(future))

            for i in range(n):
                labels[i] = results[i][0]
                boxes[i, :] = results[i][1]

            ind = boxes.sum(dim=-1) != -4
            labels = labels[ind]
            boxes = boxes[ind, :]

        # ------------------- horizontal flip
        if torch.rand(1, device=DEVICE) < FLIP_RATE:
            image = ttf.vflip(image.unsqueeze(1).transpose(1, -1).squeeze(-1))
            image = image.unsqueeze(-1).transpose(1, -1).squeeze(1)
            boxes[:, [3, 1]] = image.shape[2] - boxes[:, [1, 3]]

        # ------------------- vertical flip
        if torch.rand(1, device=DEVICE) < FLIP_RATE:
            image = ttf.hflip(image.unsqueeze(1).transpose(1, -1).squeeze(-1))
            image = image.unsqueeze(-1).transpose(1, -1).squeeze(1)
            boxes[:, [2, 0]] = image.shape[1] - boxes[:, [0, 2]]

        # ------------------- blur
        if torch.rand(1, device=DEVICE) < BLUR_RATE:
            kern: int = int(BLUR_KERNEL_TARGETS[int(torch.randint(0, len(BLUR_KERNEL_TARGETS), (1, 1)).item())].item())
            image = ttf.gaussian_blur(image.unsqueeze(1).transpose(1, -1).squeeze(-1), [kern, kern])
            image = image.unsqueeze(-1).transpose(1, -1).squeeze(1)

        # ------------------- bright
        if torch.rand(1, device=DEVICE) < BRIGHTNESS_RATE:
            # Get random brightness value for actin
            actin = (BRIGHTNESS_RANGE_ACTIN[1] - BRIGHTNESS_RANGE_ACTIN[0]) * torch.rand((1), device=DEVICE) + \
                    BRIGHTNESS_RANGE_ACTIN[0]

            # Get random brightness value for gfp
            gfp = (BRIGHTNESS_RANGE_GFP[1] - BRIGHTNESS_RANGE_GFP[0]) * torch.rand((1), device=DEVICE) + \
                  BRIGHTNESS_RANGE_GFP[0]
            val = torch.cat((gfp, actin))
            val = val.reshape(image.shape[0], 1, 1, 1)
            image.add_(val)
            # image = torch.clamp(image, 0, 1)

        # ------------------- Contrast
        if torch.rand(1, device=DEVICE) < CONTRAST_RATE or True:
            contrast_val = (CONTRAST_RANGE[1] - CONTRAST_RANGE[0]) * torch.rand((2), device=DEVICE) + \
                           CONTRAST_RANGE[0]

            for c in range(image.shape[0]):
                image[c, ..., 0] = ttf.adjust_contrast(image[[c], ..., 0], contrast_val[c].item()).squeeze(0)

        # ------------------- noise
        if torch.rand(1, device=DEVICE) < NOISE_RATE:
            image.add_(torch.rand(image.shape, device=DEVICE) * NOISE_GAMMA)
            image = torch.clamp(image, 0, 1)

        # ------------- WRAP UP
        ind = boxes[:, 0:2] < boxes[:, 2:]
        ind = torch.logical_and(ind[:, 0], ind[:, 1])

        boxes = boxes[ind, :]
        labels = labels[ind]

        data_dict['image'] = image
        data_dict['masks'] = torch.empty([0], device=DEVICE)
        data_dict['boxes'] = boxes
        data_dict['labels'] = labels

        return data_dict


@torch.jit.script
@torch.no_grad()
def mergeD_transform_3D(data_dict: Dict[str, Tensor]) -> Dict[str, Tensor]:
    DEVICE: str = str(data_dict['image'].device)

    NUL_CROP_RATE = torch.tensor(0.95, device=DEVICE)
    NUL_CROP_Z_LIMIT = torch.tensor(3, device=DEVICE)

    # Image should be in shape of [C, H, W, D]
    CROP_WIDTH = torch.tensor([300], device=DEVICE)
    CROP_HEIGHT = torch.tensor([300], device=DEVICE)
    CROP_DEPTH = torch.tensor([26], device=DEVICE)

    # AFFINE_RATE = torch.tensor(-1, device=DEVICE)
    # AFFINE_SCALE = torch.tensor((0.75, 1.25), device=DEVICE)
    # AFFINE_YAW = torch.tensor((-180, 180), device=DEVICE)
    # AFFINE_SHEAR = torch.tensor((-4, 4), device=DEVICE)

    FLIP_RATE = torch.tensor(0.5, device=DEVICE)

    # BLUR_RATE = torch.tensor(0.3, device=DEVICE)
    # BLUR_KERNEL_TARGETS = torch.tensor([3], device=DEVICE, dtype=torch.int)

    BRIGHTNESS_RATE = torch.tensor(0.5, device=DEVICE)
    BRIGHTNESS_RANGE = torch.tensor((-0.1, 0.1), device=DEVICE)

    # CONTRAST_RATE = torch.tensor(0.3, device=DEVICE)
    # CONTRAST_RANGE = torch.tensor((0.75, 2.), device=DEVICE)

    NOISE_GAMMA = torch.tensor(0.07, device=DEVICE)
    NOISE_RATE = torch.tensor(0.5, device=DEVICE)


    masks = data_dict['masks'].to(DEVICE)
    image = data_dict['image'].to(DEVICE)
    centroids = data_dict['centroids'].to(DEVICE)

    # --------- Nul Crop
    shape = data_dict['masks'].shape

    if torch.rand(1) < NUL_CROP_RATE:
        ind = torch.nonzero(masks)  # -> [I, 4] where 4 is ndims

        x_max: int = ind[:, 1].max().int().item()
        y_max: int = ind[:, 2].max().int().item()
        z_max: int = ind[:, 3].max().int().item()

        x = ind[:, 1].min().int().item()
        y = ind[:, 2].min().int().item()
        z = ind[:, 3].min().int().item() - NUL_CROP_Z_LIMIT
        z = z if z > 0 else 0

        w = x_max - x
        h = y_max - y
        d = z_max + NUL_CROP_Z_LIMIT - z
        d = d if d < shape[-1] else shape[-1]

        image = image[..., x:x + w, y:y + h, z:z + d]
        masks = masks[..., x:x + w, y:y + h, z:z + d]

    # ------------ Random Crop
    w = CROP_WIDTH if CROP_WIDTH <= image.shape[1] else torch.tensor(image.shape[1])
    h = CROP_HEIGHT if CROP_HEIGHT <= image.shape[2] else torch.tensor(image.shape[2])
    d = CROP_DEPTH if CROP_DEPTH <= image.shape[3] else torch.tensor(image.shape[3])

    if image.shape[1] > CROP_WIDTH or image.shape[2] > CROP_HEIGHT:
        ind = torch.randint(centroids.shape[0], (1, 1), dtype=torch.long, device=DEVICE)  # randomly select box
        center = centroids[ind, :].squeeze()

        x0 = center[0].sub(torch.floor(w / 2)).long().clamp(min=0, max=image.shape[2] - w.item())
        y0 = center[1].sub(torch.floor(h / 2)).long().clamp(min=0, max=image.shape[1] - h.item())
        z0 = center[2].sub(torch.floor(h / 2)).long().clamp(min=0, max=image.shape[3] - h.item())

        x1 = x0 + w
        y1 = y0 + h
        z1 = z0 + d

        image = image[:, y0.item():y1.item(), x0.item():x1.item()]

        centroids[:, 0] = centroids[:, 0] - x0
        centroids[:, 1] = centroids[:, 1] - y0
        centroids[:, 2] = centroids[:, 2] - z0

        ind_x = torch.logical_and(centroids[:, 0] >= 0, centroids[:, 2] < w)
        ind_y = torch.logical_and(centroids[:, 1] >= 0, centroids[:, 3] < h)
        ind_z = torch.logical_and(centroids[:, 2] >= 0, centroids[:, 3] < h)
        ind = torch.logical_and(ind_x, torch.logical_and(ind_y, ind_z))

        centroids = centroids[ind, :]

    # ------------------- x flip
    if torch.rand(1, device=DEVICE) < FLIP_RATE:
        image = image.flip(1)
        masks = masks.flip(1)
        centroids[:, 0] = image.shape[1] - centroids[:, 0]

    # ------------------- y flip
    if torch.rand(1, device=DEVICE) < FLIP_RATE:
        image = image.flip(2)
        masks = masks.flip(2)
        centroids[:, 1] = image.shape[2] - centroids[:, 1]

    # ------------------- z flip
    if torch.rand(1, device=DEVICE) < FLIP_RATE:
        image = image.flip(3)
        masks = masks.flip(3)
        centroids[:, 2] = image.shape[3] - centroids[:, 2]

    if torch.rand(1, device=DEVICE) < BRIGHTNESS_RATE:
        # funky looking but FAST
        val = torch.FloatTensor(image.shape[0], device=DEVICE).uniform_(BRIGHTNESS_RANGE[0], BRIGHTNESS_RANGE[1])
        image = image.add(val.reshape(image.shape[0], 1, 1, 1)).clamp(0, 1)

    if torch.rand(1, device=DEVICE) < NOISE_RATE:
        noise = torch.rand(image.shape, device=DEVICE) * NOISE_GAMMA
        image = image + noise

    return data_dict

@torch.jit.script
@torch.no_grad()
def get_centroids(masks: Tensor) -> Tensor:
    unique = torch.unique(masks)
    unique = unique[unique != 0]

    for id in unique:
        index = torch.nonzero(masks == id).mean(0)




import torch
import torch.nn as nn
from warnings import filterwarnings

filterwarnings("ignore", category=UserWarning)


@torch.jit.script
def crop(x, y):
    """
    Cropping Function to crop tensors to each other. By default only crops last 2 (in 2d) or 3 (in 3d) dimensions of
    a tensor.
    :param x: Tensor to be cropped
    :param y: Tensor by who's dimmension will crop x
    :return:
    """
    shape_x = x.shape
    shape_y = y.shape
    cropped_tensor = torch.empty(0)

    assert shape_x[1] == shape_y[1],\
        f'Inputs do not have same number of feature dimmensions: {shape_x} | {shape_y}'

    if len(shape_x) == 4:
        cropped_tensor = x[:, :, 0:shape_y[2]:1, 0:shape_y[3]:1]
    if len(shape_x) == 5:
        cropped_tensor = x[:, :, 0:shape_y[2]:1, 0:shape_y[3]:1, 0:shape_y[4]:1]

    return cropped_tensor


class Down(nn.Module):
    def __init__(self,
                 in_channels: int,
                 out_channels: int,
                 kernel: dict,
                 dilation: dict,
                 groups: dict,
                 padding=None
                 ):
        super(Down, self).__init__()
        if padding is None:
            padding = 0

        self.conv1 = nn.Conv3d(in_channels,
                                       out_channels,
                                       kernel['conv1'],
                                       dilation=dilation['conv1'],
                                       groups=groups['conv1'],
                                       padding=padding)

        self.conv2 = nn.Conv3d(out_channels,
                                       out_channels,
                                       kernel['conv2'],
                                       dilation=dilation['conv2'],
                                       groups=groups['conv2'],
                                       padding=1)

        self.batch1 = nn.BatchNorm3d(out_channels)
        self.batch2 = nn.BatchNorm3d(out_channels)
        self.relu = nn.ReLU(inplace=True)

    def forward(self, x):
        x = self.relu(self.batch1(self.conv1(x)))
        x = self.relu(self.batch2(self.conv2(x)))
        return x



import torch
import torch.nn as nn
from src.models.depreciated.StackedDilation import StackedDilation
from warnings import filterwarnings

filterwarnings("ignore", category=UserWarning)


class RDCBlock(nn.Module):
    def __init__(self, in_channels: int) -> None:
        super(RDCBlock, self).__init__()

        self.conv1x1 = nn.Conv3d(in_channels * 2, in_channels, kernel_size=1)
        self.grouped_conv = StackedDilation(in_channels, in_channels, kernel=5)
        self.activation = nn.ReLU()

        self.batch_norm_conv = nn.BatchNorm3d(in_channels)
        self.batch_norm_group = nn.BatchNorm3d(in_channels)

    def forward(self, x: torch.Tensor) -> torch.Tensor:

        x = self.activation(self.batch_norm_conv(self.conv1x1(x)))
        x = self.activation(self.batch_norm_group(self.grouped_conv(x)))

        return x



import torch
from typing import Optional
from cellpose import models
from hcat.train.dataloader import colormask_to_torch_mask
from hcat.backends.backend import Backend
from hcat import ShapeError


# DOCUMENTED

class Cellpose(Backend):
    """
    3D Hair Cell Segmentation Backend based on the cellpose algorithm:

    Stringer, Carsen, Wang, Tim, Michaelos, Michalis, & Pachitariu, Marius. (2021).
    Cellpose: A generalist algorithm for cellular segmentation. Nature Methods, 18(1), 100-106.

    """

    def __init__(self, device: Optional[str] = 'cuda'):
        """
        Initializes a cell segmentation backbone.

        .. warning::
           Initializes the cellpose model using the cellpose external website: https://www.cellpose.org/models/

        :param device: String representing pytorch device by which to evaluate the model on.
        """
        super(Cellpose, self).__init__()
        self.model = models.Cellpose(gpu=device, model_type='cyto')
        self.channels = [0, 0]

    def forward(self, image: torch.Tensor) -> torch.Tensor:
        """
        Inputs an image and outputs a probability mask of everything seen in the image.

        .. note::
           Call the module as a function to execute this method (similar to torch.nn.module).

        .. warning::
           Quirks in execution means that the image will move from gpu to cpu then back.

        .. warning:
           Will not raise an error upon failure, instead returns None and prints to standard out


        Example:

        >>> from hcat.backends.cellpose import Cellpose
        >>> import torch
        >>> backend = Cellpose()
        >>> image = torch.load('path/to/my/image.trch')
        >>> assert image.ndim == 5 # Shape should be [B, C, X, Y, Z]
        >>> masks = backend(image)


        :param image: [B, C=4, X, Y, Z] input image
        :return: [B, N, X, Y, Z] output segmentation mask where each pixel value is a cell id (0 is background)
        """

        if image.ndim != 5:
            raise ShapeError(
                f'Image ndim should be 5 with shape [B, C, X, Y, Z] not {image.ndim} with shape {image.shape}')
        b, c, x, y, z = image.shape

        if self.image_reject and self._is_image_bad(image):
            return torch.zeros((b, 0, x, y, z), device=self.device)

        masks, flows, styles, diams = self.model.eval(image.cpu().numpy(), diameter=None, channels=self.channels,
                                                      channel_axis=1, z_axis=-1, do_3D=True,
                                                      anisotropy=3, progress=False)

        masks = torch.from_numpy(masks.transpose(1, 2, 0)).unsqueeze(0)

        # Converts a mask of shape [B, C=1, X, Y, Z] where each pixel is a cell id (0 for background)
        # to a mask of shape [B, N, X, Y, Z] where N is the number of cells
        masks = colormask_to_torch_mask(masks)

        return masks.unsqueeze(0)
